<?
  # (C) 2008 Vladislav Staroselskiy
  #
  # This script parses auth log file for failed login attempts and
  # then blocks IP addresses of using ipfw
  #
  # Written and tested on FreeBSD 6.2
  #
  # How to use it:
  #
  # 1) Add the following string into appropriate part of /etc/rc.firewall file:
  // ${fwcmd} add deny ip from 'table(1)' to any via xl0
  #
  # 2) Add execution of this script into /etc/crontab file:
  // */15 * * * * root /usr/local/bin/php <path>/parse2block.php
  #
  # 3) If you use safe_mode for PHP, make sure you save this script into safe_mode_exec_dir
  # Read further information here: http://us.php.net/function.exec
  #
  # 4) Edit settings below, according to your needs.
  #
  # 5) Enjoy watching list of IPs - it will be growing fast. :)

  # ----------------
  #     Settings
  # ----------------

  $logfile="/var/log/auth.log";
  $capture_regexp="/.*Invalid\suser.*from\s(.*?)$/";
  $blockedipfile="/var/log/parse2block.txt";
  $max_incorrect_attempts=2;
  $white_list=array("192.168.0","192.168.1","192.168.2"); // Always skip this C-class subnets from checking

  $flush_command="/sbin/ipfw table 1 flush";
  $add_command="/sbin/ipfw table 1 add ";

  $debugmode=false; // set to 'true' while testing, it will simulate everything, giving a verbose output

  # ----------------
  #      Body
  # ----------------

  $blocked_ip=array();
  $dns_cache=array();
  $attempts_counter=array();
  $blocked_ip=array();

  if (is_file($blockedipfile)) {
    $h = @fopen($blockedipfile, "r");
    if ($h) {
      if ($debugmode) print "Reading data from ".$blockedipfile." file...\n";
      while (!feof($h)) {
        $s = trim(fgets($h, 256));
        if ($s!="" && preg_match("/^\d+\.\d+\.\d+\.\d+$/",$s) && array_search($s,$blocked_ip)===false) {
          $blocked_ip[] = $s;
        }
      }
      fclose($h);
    } else {
      if ($debugmode) print "ERROR: Cannot open file ".$blockedipfile." for reading\n";
    }
    if ($debugmode) print_r($blocked_ip);
  } else {
    if ($debugmode) print "WARNING: File ".$blockedipfile." doesn't exist or not readable\n";
  }

  if (is_file($logfile)) {
    $h = @fopen($logfile, "r");
    if ($h) {
      if ($debugmode) print "Reading data from ".$logfile." file...\n";
      while (!feof($h)) {
        $line = fgets($h, 256);
        if (preg_match($capture_regexp,$line,$matches)) {
          $s=$matches[1];
          if ($debugmode) print "Caught '".$s."' string";
          if (! preg_match("/^\d+\.\d+\.\d+\.\d+$/",$s)) {
            if (isset($dns_cache[$s])) {
              $ip=$dns_cache[$s];
              if ($debugmode) print ", found IP of ".$ip." in a cache";
            } else {
              $ip=gethostbyname($s);
              $dns_cache[$s]=$ip;
              if ($debugmode) print ", resolved IP address of ".$ip;
            }
          } else {
            $ip=$s;
            if ($debugmode) print ", this is an IP address";
          }
          if (preg_match("/^\d+\.\d+\.\d+\.\d+$/",$ip)) {
            if (array_search(substr($ip,0,strrpos($ip,'.')),$white_list)!==false) {
              if ($debugmode) print ", skipping IP from the white list";
            } elseif (array_search($ip,$blocked_ip)!==false) {
              if ($debugmode) print ", it is already on a blocked list";
            } else {
              if (! isset($attempts_counter[$ip])) {
                $attempts_counter[$ip]=1;
                if ($debugmode) print ", first mentioned with count=".$attempts_counter[$ip];
              } else {
                $attempts_counter[$ip]++;
                if ($debugmode) print ", recorded with count=".$attempts_counter[$ip];
              }
              if ($attempts_counter[$ip]>=$max_incorrect_attempts) {
                $blocked_ip[]=$ip;
                if ($debugmode) print ", failed attempts reached ".$max_incorrect_attempts.", IP is now BLOCKED";
              }
            }
          }
          if ($debugmode) print "\n";
        }
      }
      fclose($h);
    } else {
      if ($debugmode) print "ERROR: Cannot open file ".$logfile." for reading\n";
    }
  } else {
    if ($debugmode) print "ERROR: File ".$logfile." doesn't exist or not readable\n";
  }
  asort($blocked_ip);
  if ($debugmode) {
    print "New list of blocked IP addresses:\n--------------------------------------------\n";
    print_r($blocked_ip);
  } else {
    $h = @fopen($blockedipfile, "w+");
    if ($h) {
      exec($flush_command);
      foreach($blocked_ip as $ip) {
        fwrite($h,$ip."\n");
        exec($add_command." ".$ip);
      }
      fclose($h);
    }
  }
?>
